/*
  M2B_battleship.h

  Copyright 2009 Dhruv Batra (batradhruv -at- cmu.edu, www.ece.cmu.edu/~dbatra)
    
  This file implements the "battleship encoding" multi-label to boolean energy function transformation described in:
    
  "TRANSFORMING MULTI-LABEL HIGHER ORDER CRFS TO BOOLEAN PAIRWISE ONES", Ramalingam et. al., SIAM '09.

*/


#include <stdio.h>
//#include "graph.h"
#include "time.h"
#include <math.h>
#ifdef _MSC_VER
#pragma warning(disable: 4661)
typedef __int32 int32_t;
typedef __int64 int64_t;
typedef unsigned __int32 uint32_t;
#else
#include <stdint.h>
#endif

// row-major array access
#define ARR_RM(arr, r_ind, c_ind, ncols) (*(arr + r_ind*ncols + c_ind))
// col-major array access
#define ARR_CM(arr, r_ind, c_ind, nrows) (*(arr + c_ind*nrows + r_ind))

//#define INFY_COST 1000000000000000.0000
#define INFY_COST 100000
#define EPS 0.0001
#define NB_LABELS 2

template <typename REAL> class order2_multi
{
 public:
  
  // Default Constructor: Assign all variables to zero, and pointers to NULL. No space allocated.
  order2_multi(); 
  
  // Slightly smarter Constructor: Initialize variables. Assign pointers to pre-created locations. Space allocated for output only.
  order2_multi(uint32_t n_nodes, uint32_t n_edges, uint32_t n_labels, REAL *o2m_node_en, REAL *o2m_edge_en, uint32_t *edge_list);

  
  // Everything pre-created. No space allocated for anything.
  order2_multi(uint32_t n_nodes, uint32_t n_edges, uint32_t n_labels,	\
	       REAL *_o2m_node_en, REAL *_o2m_edge_en, uint32_t *_o2m_edge_list, \
	       REAL *_o2b_node_en, REAL *_o2b_edge_en, uint32_t *_o2b_edge_list);
  

  // Destructor. Releases all memory.
  ~order2_multi(); // Does nothing right now.
  
  
  void multi2bool();
  
 protected:
  uint32_t nm_nodes, nm_edges, nm_labels;
  uint32_t nb_nodes, nb_edges;
  
  
  REAL *o2m_node_en; // nm_labels x nm_nodes matrix
  REAL *o2m_edge_en; // (nm_labels^2) x nm_edges matrix
  uint32_t *o2m_edge_list; // 2 x nm_edges matrix
  
  
  REAL *o2b_node_en; // 2 x nb_nodes matrix
  REAL *o2b_edge_en; // 4 x nb_edges matrix 
  uint32_t *o2b_edge_list; // 2 x nb_edges matrix

public:  
  REAL offset;
};


template <typename REAL>
inline order2_multi<REAL>::order2_multi() :   
  nm_nodes(0), nm_edges(0), nm_labels(0), nb_nodes(0), nb_edges(0), 
  o2m_node_en(NULL), o2m_edge_en(NULL), o2m_edge_list(NULL),
  o2b_node_en(NULL), o2b_edge_en(NULL), o2b_edge_list(NULL),
  offset(0)
{}


template <typename REAL>
inline order2_multi<REAL>::order2_multi(uint32_t n_nodes, uint32_t n_edges, uint32_t n_labels, \
					REAL *_o2m_node_en, REAL *_o2m_edge_en, uint32_t *_o2m_edge_list, \
					REAL *_o2b_node_en, REAL *_o2b_edge_en, uint32_t *_o2b_edge_list) :
  nm_nodes(n_nodes), nm_edges(n_edges), nm_labels(n_labels), nb_nodes(n_nodes*(n_labels-1)), nb_edges(n_nodes*(n_labels-2)+(n_labels-1)*(n_labels-1)*n_edges), 
  o2m_node_en(_o2m_node_en), o2m_edge_en(_o2m_edge_en), o2m_edge_list(_o2m_edge_list),
  o2b_node_en(_o2b_node_en), o2b_edge_en(_o2b_edge_en), o2b_edge_list(_o2b_edge_list),
  offset(0)
{}


template <typename REAL>
inline order2_multi<REAL>::~order2_multi() 
{}


template <typename REAL>
inline void order2_multi<REAL>::multi2bool()
{
  uint32_t id,id1,id2;
  REAL sCost,tCost,pCost,tmpCost,tmpCost2;
  int n1,n2,l1,l2,e1;
  int counter;
  int i,j,tmp;
  
  // encode unary terms
  counter = 0;
  for(n1=0; n1<nm_nodes; n1++)
    {
      //fscanf(fp,"%lf\n",&tCost);
      tCost = ARR_CM(o2m_node_en,0,n1,nm_labels);
      
      for(int l1=0; l1<(nm_labels-1); l1++)
	{
	  //g -> add_node(); 
	  id = n1*(nm_labels-1) + l1;/* Id of the node */
	  
	  //fscanf(fp,"%lf\n",&sCost);
	  sCost = ARR_CM(o2m_node_en,l1+1,n1,nm_labels);
	  
	  //g -> add_tweights(id,sCost,tCost);		
	  ARR_CM(o2b_node_en,0,id,NB_LABELS) += tCost;
	  ARR_CM(o2b_node_en,1,id,NB_LABELS) += sCost;
	  
	  if (l1>0)
	    offset -= tCost;
	  tCost=sCost;
	  
	  if (l1>0)
	    {
	      //g -> add_edge(id-1,id,INFY_COST,0);
	      ARR_CM(o2b_edge_list,0,counter,2) = id-1;
	      ARR_CM(o2b_edge_list,1,counter,2) = id;
	      
	      ARR_CM(o2b_edge_en,2,counter,4) = INFY_COST; // Can't understand what's happening. 
	      //ARR_CM(o2b_edge_en,1,counter,4) = INFY_COST; // It should be the first line, but second line gives the right ans.
	      counter++;
	    }
	  
	}
    }
  
  tmp = nm_labels*nm_labels;

  for (e1=0; e1<nm_edges; e1++)
    {
      //fscanf(fp,"%d,%d\n",&id1,&id2);
      //printf("%d,%d\n",id1,id2);
      n1 = ARR_CM(o2m_edge_list,0,e1,2);
      n2 = ARR_CM(o2m_edge_list,1,e1,2);

      for (l1=0; l1<(nm_labels-1); l1++)
	{
	  id1 = n1*(nm_labels-1)+l1;
	  
	  for (l2=0; l2<(nm_labels-1); l2++)
	    {
	      id2 = n2*(nm_labels-1)+l2;
	      
	      //pCost=(*(edgeCost+l1*nm_labels+l2)) -(*(edgeCost+(l1+1)*nm_labels+l2))-(*(edgeCost+l1*nm_labels+l2+1))+(*(edgeCost+(l1+1)*nm_labels+l2+1));
	      pCost = ARR_CM(o2m_edge_en, l2*nm_labels+l1, e1, tmp) - ARR_CM(o2m_edge_en, (l2+1)*nm_labels+l1, e1, tmp) \
		- ARR_CM(o2m_edge_en, l2*nm_labels+l1+1, e1, tmp) + ARR_CM(o2m_edge_en, (l2+1)*nm_labels+l1+1, e1, tmp);
	      
// 	      if (pCost > EPS) {
// 		printf("Warning:Multilabel term %d not submodular!: Alpha = %f\n", (int)e1, (float) pCost);
// 		for (i=0; i<nm_labels; i++) {
// 		  for (j=0; j<nm_labels; j++)
// 		    printf("%f ",(float) ARR_CM(o2m_edge_en,j*nm_labels+i,e1,tmp));
// 		  printf("\n");
// 		}
// 	      }
	      

	      //g -> add_edge(n1*(nm_labels-1)+l1,n2*(nm_labels-1)+l2,-pCost,0);
	      ARR_CM(o2b_edge_list,0,counter,2) = id1;
	      ARR_CM(o2b_edge_list,1,counter,2) = id2;
	      
	      ARR_CM(o2b_edge_en,2,counter,4) = -pCost; // Can't understand what's happening. 
	      //ARR_CM(o2b_edge_en,1,counter,4) = -pCost; // It should be the first line, but second line gives the right ans.
	      counter++;
	      
	      //g -> add_tweights(n2*(nm_labels-1)+l2,0,-pCost);
	      ARR_CM(o2b_node_en, 0, id2, NB_LABELS) += (-pCost);

	      offset += pCost;
	    }
	  
	  // Adding unary costs.
	  //g -> add_tweights(n1*(nm_labels-1)+l1,(*(edgeCost+(l1+1)*nm_labels)),*(edgeCost+(l1)*nm_labels));
	  tmpCost = ARR_CM(o2m_edge_en, l1, e1, tmp);
	  ARR_CM(o2b_node_en, 0, id1, NB_LABELS) += tmpCost;
	  ARR_CM(o2b_node_en, 1, id1, NB_LABELS) += ARR_CM(o2m_edge_en, l1+1, e1, tmp);
		 
	  //offset -= (*(edgeCost+(l1)*nm_labels));
	  offset -= tmpCost;
	    
	  //g -> add_tweights(n2*(nm_labels-1)+l1,*(edgeCost+l1+1),*(edgeCost+l1));
	  id = n2*(nm_labels-1)+l1;
	  tmpCost = ARR_CM(o2m_edge_en, l1*nm_labels, e1, tmp);
	  ARR_CM(o2b_node_en, 0, id, NB_LABELS) += tmpCost;
	  ARR_CM(o2b_node_en, 1, id, NB_LABELS) += ARR_CM(o2m_edge_en, (l1+1)*nm_labels, e1, tmp);
	  
	  //offset -= (*(edgeCost+l1));
	  offset -= tmpCost;
	}
      
      // Adding unary costs.
      //offset += *(edgeCost);
      offset += ARR_CM(o2m_edge_en,0,e1,tmp);
      
    }
}
